# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log as logging

from magnum.common import exception
from magnum.common import keystone
from magnum.common import neutron
from magnum.drivers.heat import template_def

CONF = cfg.CONF


LOG = logging.getLogger(__name__)


"""kubernetes ports """
KUBE_SECURE_PORT = '6443'
KUBE_INSECURE_PORT = '8080'


class K8sApiAddressOutputMapping(template_def.OutputMapping):

    def set_output(self, stack, cluster_template, cluster):
        if self.cluster_attr is None:
            return

        output_value = self.get_output_value(stack, cluster)
        if output_value is not None:
            # TODO(yuanying): port number is hardcoded, this will be fix
            protocol = 'https'
            port = KUBE_SECURE_PORT
            if cluster_template.tls_disabled:
                protocol = 'http'
                port = KUBE_INSECURE_PORT

            params = {
                'protocol': protocol,
                'address': output_value,
                'port': port,
            }
            value = "%(protocol)s://%(address)s:%(port)s" % params
            setattr(cluster, self.cluster_attr, value)


class ServerAddressOutputMapping(template_def.NodeGroupOutputMapping):

    public_ip_output_key = None
    private_ip_output_key = None

    def __init__(self, dummy_arg, nodegroup_attr=None, nodegroup_uuid=None):
        self.nodegroup_attr = nodegroup_attr
        self.nodegroup_uuid = nodegroup_uuid
        self.heat_output = self.public_ip_output_key
        self.is_stack_param = False

    def set_output(self, stack, cluster_template, cluster):
        if not cluster.floating_ip_enabled:
            self.heat_output = self.private_ip_output_key

        LOG.debug("Using heat_output: %s", self.heat_output)
        super(ServerAddressOutputMapping,
              self).set_output(stack, cluster_template, cluster)


class MasterAddressOutputMapping(ServerAddressOutputMapping):
    public_ip_output_key = 'kube_masters'
    private_ip_output_key = 'kube_masters_private'


class NodeAddressOutputMapping(ServerAddressOutputMapping):
    public_ip_output_key = 'kube_minions'
    private_ip_output_key = 'kube_minions_private'


class K8sTemplateDefinition(template_def.BaseTemplateDefinition):
    """Base Kubernetes template."""

    def __init__(self):
        super(K8sTemplateDefinition, self).__init__()
        self.add_parameter('external_network',
                           cluster_template_attr='external_network_id',
                           required=True)
        self.add_parameter('fixed_network',
                           cluster_attr='fixed_network')
        self.add_parameter('fixed_subnet',
                           cluster_attr='fixed_subnet')
        self.add_parameter('network_driver',
                           cluster_template_attr='network_driver')
        self.add_parameter('volume_driver',
                           cluster_template_attr='volume_driver')
        self.add_parameter('tls_disabled',
                           cluster_template_attr='tls_disabled',
                           required=True)
        self.add_parameter('registry_enabled',
                           cluster_template_attr='registry_enabled')
        self.add_parameter('cluster_uuid',
                           cluster_attr='uuid',
                           param_type=str)
        self.add_parameter('insecure_registry_url',
                           cluster_template_attr='insecure_registry')
        self.add_parameter('kube_version',
                           cluster_attr='coe_version')

        self.add_output('api_address',
                        cluster_attr='api_address',
                        mapping_type=K8sApiAddressOutputMapping)
        self.add_output('kube_minions_private',
                        cluster_attr=None)
        self.add_output('kube_masters_private',
                        cluster_attr=None)
        self.default_subnet_cidr = '10.0.0.0/24'

    def get_nodegroup_param_maps(self, master_params=None, worker_params=None):
        master_params = master_params or dict()
        worker_params = worker_params or dict()
        master_params.update({
            'master_flavor': 'flavor_id',
            'master_image': 'image_id',
            'master_role': 'role',
            'master_nodegroup_name': 'name',
            'docker_volume_size': 'docker_volume_size'
        })
        worker_params.update({
            'number_of_minions': 'node_count',
            'minion_flavor': 'flavor_id',
            'minion_image': 'image_id',
            'worker_role': 'role',
            'worker_nodegroup_name': 'name',
            'docker_volume_size': 'docker_volume_size'
        })
        return super(
            K8sTemplateDefinition, self).get_nodegroup_param_maps(
                master_params=master_params, worker_params=worker_params)

    def update_outputs(self, stack, cluster_template, cluster,
                       nodegroups=None):
        nodegroups = nodegroups or [cluster.default_ng_worker,
                                    cluster.default_ng_master]
        for nodegroup in nodegroups:
            if nodegroup.role == 'master':
                self.add_output('kube_masters',
                                nodegroup_attr='node_addresses',
                                nodegroup_uuid=nodegroup.uuid,
                                mapping_type=MasterAddressOutputMapping)
            else:
                self.add_output('kube_minions',
                                nodegroup_attr='node_addresses',
                                nodegroup_uuid=nodegroup.uuid,
                                mapping_type=NodeAddressOutputMapping)
                self.add_output(
                    'number_of_minions', nodegroup_attr='node_count',
                    nodegroup_uuid=nodegroup.uuid,
                    mapping_type=template_def.NodeGroupOutputMapping,
                    is_stack_param=True)
        super(K8sTemplateDefinition,
              self).update_outputs(stack, cluster_template, cluster,
                                   nodegroups=nodegroups)

    def get_net_params(self, context, cluster_template, cluster):
        extra_params = dict()
        # NOTE(lxkong): Convert external network name to UUID, the template
        # field name is confused. If external_network_id is not specified in
        # cluster template use 'public' as the default value, which is the same
        # with the heat template default value as before.
        external_network = cluster_template.external_network_id
        ext_net_id = neutron.get_external_network_id(context, external_network)
        extra_params['external_network'] = ext_net_id

        # NOTE(brtknr): Convert fixed network UUID to name if the given network
        # name is UUID like because OpenStack Cloud Controller Manager only
        # accepts a name as an argument to internal-network-name in the
        # cloud-config file provided to it. The default fixed network name is
        # the same as that defined in the heat template.
        fixed_network = cluster.fixed_network
        net_name = neutron.get_fixed_network_name(context, fixed_network)
        if net_name:
            extra_params['fixed_network_name'] = net_name
        else:
            extra_params['fixed_network_name'] = cluster.name

        if cluster.labels.get('fixed_subnet_cidr'):
            extra_params['fixed_subnet_cidr'] = cluster.labels.get(
                'fixed_subnet_cidr')

        # NOTE(brtknr): Convert fixed subnet name to UUID. If fixed_subnet
        # is not specified in cluster template use 'private' as the default
        # value, which is the same as the heat template default value.
        fixed_subnet = cluster.fixed_subnet
        subnet_id = neutron.get_fixed_subnet_id(context, fixed_subnet)
        if subnet_id:
            extra_params['fixed_subnet'] = subnet_id
            # NOTE(flwang): If a fixed subnet is given, then the label
            # fixed_subnet_cidr should be updated to reflect the correct
            # setting.
            extra_params['fixed_subnet_cidr'] = neutron.get_subnet(
                context, subnet_id, "id", "cidr")

        if cluster_template.no_proxy:
            extra_params["no_proxy"] = (
                cluster_template.no_proxy + "," + (
                    extra_params.get('fixed_subnet_cidr') or
                    self.default_subnet_cidr))

        return extra_params

    def get_params(self, context, cluster_template, cluster, **kwargs):
        extra_params = kwargs.pop('extra_params', {})

        extra_params['discovery_url'] = self.get_discovery_url(cluster)
        osc = self.get_osc(context)
        # NOTE: Sometimes, version discovery fails when Magnum cannot talk to
        # Keystone via specified magnum_client.endpoint_type intended for
        # cluster instances either because it is not unreachable from the
        # controller or CA certs are missing for TLS enabled interface and the
        # returned auth_url may not be suffixed with /v1 in which case append
        # the url with the suffix so that instances can still talk to Magnum.
        magnum_url = osc.magnum_url()
        extra_params['magnum_url'] = magnum_url + ('' if
                                                   magnum_url.endswith('/v1')
                                                   else '/v1')

        if cluster_template.tls_disabled:
            extra_params['loadbalancing_protocol'] = 'HTTP'
            extra_params['kubernetes_port'] = 8080

        extra_params['octavia_enabled'] = keystone.is_octavia_enabled()

        net_params = self.get_net_params(context, cluster_template, cluster)
        extra_params.update(net_params)

        label_list = ['flannel_network_cidr', 'flannel_backend',
                      'flannel_network_subnetlen',
                      'system_pods_initial_delay',
                      'system_pods_timeout',
                      'admission_control_list',
                      'prometheus_monitoring',
                      'grafana_admin_passwd',
                      'kube_dashboard_enabled',
                      'etcd_volume_size',
                      'cert_manager_api',
                      'ingress_controller_role',
                      'octavia_ingress_controller_tag',
                      'kubelet_options',
                      'kubeapi_options',
                      'kubeproxy_options',
                      'kubecontroller_options',
                      'kubescheduler_options',
                      'influx_grafana_dashboard_enabled',
                      'master_lb_allowed_cidrs']

        labels = self._get_relevant_labels(cluster, kwargs)

        for label in label_list:
            extra_params[label] = labels.get(label)

        ingress_controller = cluster.labels.get('ingress_controller',
                                                '').lower()
        if (ingress_controller == 'octavia'
                and not extra_params['octavia_enabled']):
            raise exception.InvalidParameterValue(
                'Octavia service needs to be deployed for octavia ingress '
                'controller.')
        extra_params["ingress_controller"] = ingress_controller

        cluser_ip_range = cluster.labels.get('service_cluster_ip_range')
        if cluser_ip_range:
            extra_params['portal_network_cidr'] = cluser_ip_range

        if cluster_template.registry_enabled:
            extra_params['swift_region'] = CONF.docker_registry.swift_region
            extra_params['registry_container'] = (
                CONF.docker_registry.swift_registry_container)

        kube_tag = (labels.get("kube_tag") or
                    cluster_template.labels.get("kube_tag"))
        if kube_tag:
            extra_params['kube_version'] = kube_tag
            extra_params['master_kube_tag'] = kube_tag
            extra_params['minion_kube_tag'] = kube_tag

        self._set_master_lb_allowed_cidrs(context, cluster, extra_params)

        return super(K8sTemplateDefinition,
                     self).get_params(context, cluster_template, cluster,
                                      extra_params=extra_params,
                                      **kwargs)

    def _set_master_lb_allowed_cidrs(self, context, cluster, extra_params):
        if extra_params.get("master_lb_allowed_cidrs"):
            subnet_cidr = (cluster.labels.get("fixed_subnet_cidr") or
                           self.default_subnet_cidr)
            if extra_params.get("fixed_subnet"):
                subnet_cidr = neutron.get_subnet(context,
                                                 extra_params["fixed_subnet"],
                                                 "id", "cidr")
            extra_params["master_lb_allowed_cidrs"] += "," + subnet_cidr

    def get_scale_params(self, context, cluster, node_count,
                         scale_manager=None, nodes_to_remove=None):
        scale_params = dict()
        if nodes_to_remove:
            scale_params['minions_to_remove'] = nodes_to_remove
        if scale_manager:
            hosts = self.get_output('kube_minions_private')
            scale_params['minions_to_remove'] = (
                scale_manager.get_removal_nodes(hosts))
        scale_params['number_of_minions'] = node_count
        return scale_params



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import abc
import functools
import inspect
import time
import types


def def_method(f, *args, **kwargs):
    @functools.wraps(f)
    def new_method(self):
        return f(self, *args, **kwargs)
    return new_method


def parameterized_class(cls):
    """A class decorator for running parameterized test cases.

    Mark your class with @parameterized_class.
    Mark your test cases with @parameterized.
    """
    test_functions = inspect.getmembers(cls, predicate=inspect.ismethod)
    for (name, f) in test_functions:
        if name.startswith('test_') and not hasattr(f, '_test_data'):
            continue

        # remove the original test function from the class
        delattr(cls, name)

        # add a new test function to the class for each entry in f._test_data
        for tag, args in f._test_data.items():
            new_name = "{0}_{1}".format(f.__name__, tag)
            if hasattr(cls, new_name):
                raise Exception(
                    "Parameterized test case '{0}.{1}' created from '{0}.{2}' "
                    "already exists".format(cls.__name__, new_name, name))

            # Using `def new_method(self): f(self, **args)` is not sufficient
            # (all new_methods use the same args value due to late binding).
            # Instead, use this factory function.
            new_method = def_method(f, **args)

            # To add a method to a class, available for all instances:
            #   MyClass.method = types.MethodType(f, None, MyClass)
            setattr(cls, new_name, types.MethodType(new_method, None, cls))
    return cls


def parameterized(data):
    """A function decorator for parameterized test cases.

    Example:

        @parameterized({
            'zero': dict(val=0),
            'one': dict(val=1),
        })
        def test_val(self, val):
            self.assertEqual(val, self.get_val())

    The above will generate two test cases:
        `test_val_zero` which runs with val=0
        `test_val_one` which runs with val=1

    :param data: A dictionary that looks like {tag: {arg1: val1, ...}}
    """
    def wrapped(f):
        f._test_data = data
        return f
    return wrapped


def wait_for_condition(condition, interval=1, timeout=40):
    start_time = time.time()
    end_time = time.time() + timeout
    while time.time() < end_time:
        result = condition()
        if result:
            return result
        time.sleep(interval)
    raise Exception(("Timed out after %s seconds.  Started " +
                    "on %s and ended on %s") % (timeout, start_time, end_time))


def memoized(func):
    """A decorator to cache function's return value"""
    cache = {}

    @functools.wraps(func)
    def wrapper(*args):
        if not isinstance(args, abc.Hashable):
            # args is not cacheable. just call the function.
            return func(*args)
        if args in cache:
            return cache[args]
        else:
            value = func(*args)
            cache[args] = value
            return value
    return wrapper



# Copyright 2016 - Fujitsu, Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from oslo_concurrency import processutils

from magnum.cmd import api
from magnum.tests import base


# NOTE(hieulq): need to mock MagnumObject, otherwise other test cases
# will be failed because of setting wrong ovo indirection api
@mock.patch('magnum.objects.base.MagnumObject')
class TestMagnumAPI(base.TestCase):

    @mock.patch('werkzeug.serving.run_simple')
    @mock.patch.object(api, 'api_app')
    @mock.patch('magnum.common.service.prepare_service')
    def test_api_http(self, mock_prep, mock_app, mock_run, mock_base):
        api.main()

        app = mock_app.load_app.return_value
        mock_prep.assert_called_once_with(mock.ANY)
        mock_app.load_app.assert_called_once_with()
        workers = processutils.get_worker_count()
        mock_run.assert_called_once_with(base.CONF.api.host,
                                         base.CONF.api.port,
                                         app, processes=workers,
                                         ssl_context=None)

    @mock.patch('werkzeug.serving.run_simple')
    @mock.patch.object(api, 'api_app')
    @mock.patch('magnum.common.service.prepare_service')
    def test_api_http_config_workers(self, mock_prep, mock_app,
                                     mock_run, mock_base):
        fake_workers = 8
        self.config(workers=fake_workers, group='api')
        api.main()

        app = mock_app.load_app.return_value
        mock_prep.assert_called_once_with(mock.ANY)
        mock_app.load_app.assert_called_once_with()
        mock_run.assert_called_once_with(base.CONF.api.host,
                                         base.CONF.api.port,
                                         app, processes=fake_workers,
                                         ssl_context=None)

    @mock.patch('os.path.exists')
    @mock.patch('werkzeug.serving.run_simple')
    @mock.patch.object(api, 'api_app')
    @mock.patch('magnum.common.service.prepare_service')
    def test_api_https_no_cert(self, mock_prep, mock_app, mock_run,
                               mock_exist, mock_base):
        self.config(enabled_ssl=True,
                    ssl_cert_file='tmp_crt',
                    group='api')
        mock_exist.return_value = False

        self.assertRaises(RuntimeError, api.main)
        mock_prep.assert_called_once_with(mock.ANY)
        mock_app.load_app.assert_called_once_with()
        mock_run.assert_not_called()
        mock_exist.assert_called_once_with('tmp_crt')

    @mock.patch('os.path.exists')
    @mock.patch('werkzeug.serving.run_simple')
    @mock.patch.object(api, 'api_app')
    @mock.patch('magnum.common.service.prepare_service')
    def test_api_https_no_key(self, mock_prep, mock_app, mock_run,
                              mock_exist, mock_base):
        self.config(enabled_ssl=True,
                    ssl_cert_file='tmp_crt',
                    ssl_key_file='tmp_key',
                    group='api')
        mock_exist.side_effect = [True, False]

        self.assertRaises(RuntimeError, api.main)
        mock_prep.assert_called_once_with(mock.ANY)
        mock_app.load_app.assert_called_once_with()
        mock_run.assert_not_called()
        mock_exist.assert_has_calls([mock.call('tmp_crt'),
                                     mock.call('tmp_key')])

    @mock.patch('os.path.exists')
    @mock.patch('werkzeug.serving.run_simple')
    @mock.patch.object(api, 'api_app')
    @mock.patch('magnum.common.service.prepare_service')
    def test_api_https(self, mock_prep, mock_app, mock_run,
                       mock_exist, mock_base):
        self.config(enabled_ssl=True,
                    ssl_cert_file='tmp_crt',
                    ssl_key_file='tmp_key',
                    group='api')
        mock_exist.side_effect = [True, True]

        api.main()

        app = mock_app.load_app.return_value
        mock_prep.assert_called_once_with(mock.ANY)
        mock_app.load_app.assert_called_once_with()
        mock_exist.assert_has_calls([mock.call('tmp_crt'),
                                     mock.call('tmp_key')])
        workers = processutils.get_worker_count()
        mock_run.assert_called_once_with(base.CONF.api.host,
                                         base.CONF.api.port, app,
                                         processes=workers,
                                         ssl_context=('tmp_crt', 'tmp_key'))



# Copyright 2012 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
from oslo_policy import _parser

from magnum.common import policy as magnum_policy
import magnum.conf

CONF = magnum.conf.CONF


class PolicyFixture(fixtures.Fixture):

    def _setUp(self):
        CONF(args=[], project='magnum')
        magnum_policy._ENFORCER = None
        self.addCleanup(magnum_policy.init().clear)

    def set_rules(self, rules):
        policy = magnum_policy._ENFORCER
        policy.set_rules({k: _parser.parse_rule(v)
                          for k, v in rules.items()})



# Copyright 2015 Rackspace  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import contextlib

import docker
from docker.utils import utils

from magnum.conductor.handlers.common import cert_manager
from magnum.conductor import utils as conductor_utils
import magnum.conf


CONF = magnum.conf.CONF


def parse_docker_image(image):
    image_parts = image.split(':', 1)

    image_repo = image_parts[0]
    image_tag = None

    if len(image_parts) > 1:
        image_tag = image_parts[1]

    return image_repo, image_tag


def is_docker_library_version_atleast(version):
    if utils.compare_version(docker.version, version) <= 0:
        return True
    return False


def is_docker_api_version_atleast(docker, version):
    if utils.compare_version(docker.version()['ApiVersion'], version) <= 0:
        return True
    return False


@contextlib.contextmanager
def docker_for_cluster(context, cluster):
    cluster_template = conductor_utils.retrieve_cluster_template(
        context, cluster)

    ca_cert, magnum_key, magnum_cert = None, None, None
    client_kwargs = dict()
    if not cluster_template.tls_disabled:
        (ca_cert, magnum_key,
         magnum_cert) = cert_manager.create_client_files(cluster, context)
        client_kwargs['ca_cert'] = ca_cert.name
        client_kwargs['client_key'] = magnum_key.name
        client_kwargs['client_cert'] = magnum_cert.name

    yield DockerHTTPClient(
        cluster.api_address,
        CONF.docker.docker_remote_api_version,
        CONF.docker.default_timeout,
        **client_kwargs
    )

    if ca_cert:
        ca_cert.close()
    if magnum_key:
        magnum_key.close()
    if magnum_cert:
        magnum_cert.close()


class DockerHTTPClient(docker.APIClient):
    def __init__(self, url='unix://var/run/docker.sock',
                 ver=CONF.docker.docker_remote_api_version,
                 timeout=CONF.docker.default_timeout,
                 ca_cert=None,
                 client_key=None,
                 client_cert=None):

        if ca_cert and client_key and client_cert:
            ssl_config = docker.tls.TLSConfig(
                client_cert=(client_cert, client_key),
                verify=ca_cert,
                assert_hostname=False,
            )
        else:
            ssl_config = False

        super(DockerHTTPClient, self).__init__(
            base_url=url,
            version=ver,
            timeout=timeout,
            tls=ssl_config
        )

    def list_instances(self, inspect=False):
        res = []
        for container in self.containers(all=True):
            info = self.inspect_container(container['Id'])
            if not info:
                continue
            if inspect:
                res.append(info)
            else:
                res.append(info['Config'].get('Hostname'))
        return res



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from eventlet.green import threading
from oslo_context import context

from magnum.common import policy

import magnum.conf

CONF = magnum.conf.CONF


class RequestContext(context.RequestContext):
    """Extends security contexts from the OpenStack common library."""

    def __init__(self, auth_token=None, auth_url=None, domain_id=None,
                 domain_name=None, user_name=None, user_id=None,
                 user_domain_name=None, user_domain_id=None,
                 project_name=None, project_id=None, roles=None,
                 is_admin=None, read_only=False, show_deleted=False,
                 request_id=None, trust_id=None, auth_token_info=None,
                 all_tenants=False, password=None, **kwargs):
        """Stores several additional request parameters:

        :param domain_id: The ID of the domain.
        :param domain_name: The name of the domain.
        :param user_domain_id: The ID of the domain to
                               authenticate a user against.
        :param user_domain_name: The name of the domain to
                                 authenticate a user against.

        """
        super(RequestContext, self).__init__(auth_token=auth_token,
                                             user_id=user_name,
                                             project_id=project_name,
                                             is_admin=is_admin,
                                             read_only=read_only,
                                             show_deleted=show_deleted,
                                             request_id=request_id,
                                             roles=roles)

        self.user_name = user_name
        self.user_id = user_id
        self.project_name = project_name
        self.project_id = project_id
        self.domain_id = domain_id
        self.domain_name = domain_name
        self.user_domain_id = user_domain_id
        self.user_domain_name = user_domain_name
        self.auth_url = auth_url
        self.auth_token_info = auth_token_info
        self.trust_id = trust_id
        self.all_tenants = all_tenants
        self.password = password
        if is_admin is None:
            self.is_admin = policy.check_is_admin(self)
        else:
            self.is_admin = is_admin

    def to_dict(self):
        value = super(RequestContext, self).to_dict()
        value.update({'auth_token': self.auth_token,
                      'auth_url': self.auth_url,
                      'domain_id': self.domain_id,
                      'domain_name': self.domain_name,
                      'user_domain_id': self.user_domain_id,
                      'user_domain_name': self.user_domain_name,
                      'user_name': self.user_name,
                      'user_id': self.user_id,
                      'project_name': self.project_name,
                      'project_id': self.project_id,
                      'is_admin': self.is_admin,
                      'read_only': self.read_only,
                      'roles': self.roles,
                      'show_deleted': self.show_deleted,
                      'request_id': self.request_id,
                      'trust_id': self.trust_id,
                      'auth_token_info': self.auth_token_info,
                      'password': self.password,
                      'all_tenants': self.all_tenants})
        return value

    @classmethod
    def from_dict(cls, values):
        return cls(**values)


def make_context(*args, **kwargs):
    return RequestContext(*args, **kwargs)


def make_admin_context(show_deleted=False, all_tenants=False):
    """Create an administrator context.

    :param show_deleted: if True, will show deleted items when query db
    """
    context = RequestContext(user_id=None,
                             project=None,
                             is_admin=True,
                             show_deleted=show_deleted,
                             all_tenants=all_tenants)
    return context


def make_cluster_context(cluster, show_deleted=False):
    """Create a user context based on a cluster's stored Keystone trust.

    :param cluster: the cluster supplying the Keystone trust to use
    :param show_deleted: if True, will show deleted items when query db
    """
    context = RequestContext(user_name=cluster.trustee_username,
                             password=cluster.trustee_password,
                             trust_id=cluster.trust_id,
                             show_deleted=show_deleted,
                             user_domain_id=CONF.trust.trustee_domain_id,
                             user_domain_name=CONF.trust.trustee_domain_name)
    return context


_CTX_STORE = threading.local()
_CTX_KEY = 'current_ctx'


def has_ctx():
    return hasattr(_CTX_STORE, _CTX_KEY)


def ctx():
    return getattr(_CTX_STORE, _CTX_KEY)


def set_ctx(new_ctx):
    if not new_ctx and has_ctx():
        delattr(_CTX_STORE, _CTX_KEY)
        if hasattr(context._request_store, 'context'):
            delattr(context._request_store, 'context')

    if new_ctx:
        setattr(_CTX_STORE, _CTX_KEY, new_ctx)
        setattr(context._request_store, 'context', new_ctx)


def get_admin_context(read_deleted="no"):
    # NOTE(tovin07): This method should only be used when an admin context is
    # necessary for the entirety of the context lifetime.
    return RequestContext(user_id=None,
                          project_id=None,
                          is_admin=True,
                          read_deleted=read_deleted,
                          overwrite=False)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from magnum import objects
from magnum.tests.unit.db import base
from magnum.tests.unit.db import utils


class TestMagnumServiceObject(base.DbTestCase):

    def setUp(self):
        super(TestMagnumServiceObject, self).setUp()
        self.fake_magnum_service = utils.get_test_magnum_service()

    def test_get_by_host_and_binary(self):
        with mock.patch.object(self.dbapi,
                               'get_magnum_service_by_host_and_binary',
                               autospec=True) as mock_get_magnum_service:
            mock_get_magnum_service.return_value = self.fake_magnum_service
            ms = objects.MagnumService.get_by_host_and_binary(self.context,
                                                              'fake-host',
                                                              'fake-bin')
            mock_get_magnum_service.assert_called_once_with('fake-host',
                                                            'fake-bin')
            self.assertEqual(self.context, ms._context)

    def test_get_by_host_and_binary_no_service(self):
        with mock.patch.object(self.dbapi, 'create_magnum_service',
                               autospec=True) as mock_get_magnum_service:
            mock_get_magnum_service.return_value = None
            ms = objects.MagnumService.get_by_host_and_binary(self.context,
                                                              'fake-host',
                                                              'fake-bin')

            self.assertIsNone(ms)

    def test_create(self):
        with mock.patch.object(self.dbapi, 'create_magnum_service',
                               autospec=True) as mock_create_magnum_service:
            mock_create_magnum_service.return_value = self.fake_magnum_service
            ms_dict = {'host': 'fakehost', 'binary': 'fake-bin'}
            ms = objects.MagnumService(self.context, **ms_dict)
            ms.create(self.context)
            mock_create_magnum_service.assert_called_once_with(ms_dict)

    def test_destroy(self):
        with mock.patch.object(self.dbapi,
                               'get_magnum_service_by_host_and_binary',
                               autospec=True) as mock_get_magnum_service:
            mock_get_magnum_service.return_value = self.fake_magnum_service
            with mock.patch.object(self.dbapi,
                                   'destroy_magnum_service',
                                   autospec=True) as mock_destroy_ms:
                ms = objects.MagnumService.get_by_host_and_binary(
                    self.context, 'fake-host', 'fake-bin')
                ms.destroy()
                mock_get_magnum_service.assert_called_once_with(
                    'fake-host', 'fake-bin')
                mock_destroy_ms.assert_called_once_with(
                    self.fake_magnum_service['id'])
                self.assertEqual(self.context, ms._context)

    def test_save(self):
        with mock.patch.object(self.dbapi,
                               'get_magnum_service_by_host_and_binary',
                               autospec=True) as mock_get_magnum_service:
            mock_get_magnum_service.return_value = self.fake_magnum_service
            with mock.patch.object(self.dbapi,
                                   'update_magnum_service',
                                   autospec=True) as mock_update_ms:
                ms = objects.MagnumService.get_by_host_and_binary(
                    self.context, 'fake-host', 'fake-bin')
                ms.disabled = True
                ms.save()
                mock_get_magnum_service.assert_called_once_with(
                    'fake-host', 'fake-bin')
                mock_update_ms.assert_called_once_with(
                    self.fake_magnum_service['id'], {'disabled': True})
                self.assertEqual(self.context, ms._context)

    def test_report_state_up(self):
        with mock.patch.object(self.dbapi,
                               'get_magnum_service_by_host_and_binary',
                               autospec=True) as mock_get_magnum_service:
            mock_get_magnum_service.return_value = self.fake_magnum_service
            with mock.patch.object(self.dbapi,
                                   'update_magnum_service',
                                   autospec=True) as mock_update_ms:
                ms = objects.MagnumService.get_by_host_and_binary(
                    self.context, 'fake-host', 'fake-bin')
                last_report_count = self.fake_magnum_service['report_count']
                ms.report_state_up()
                mock_get_magnum_service.assert_called_once_with(
                    'fake-host', 'fake-bin')
                self.assertEqual(self.context, ms._context)
                mock_update_ms.assert_called_once_with(
                    self.fake_magnum_service['id'],
                    {'report_count': last_report_count + 1})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Add docker storage driver column

Revision ID: a1136d335540
Revises: d072f58ab240
Create Date: 2016-03-07 19:00:28.738486

"""

# revision identifiers, used by Alembic.
revision = 'a1136d335540'
down_revision = 'd072f58ab240'

from alembic import op  # noqa: E402
import sqlalchemy as sa  # noqa: E402


docker_storage_driver_enum = sa.Enum('devicemapper', 'overlay',
                                     name='docker_storage_driver')


def upgrade():
    docker_storage_driver_enum.create(op.get_bind(), checkfirst=True)
    op.add_column('baymodel', sa.Column('docker_storage_driver',
                                        docker_storage_driver_enum,
                                        nullable=True))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from magnum.conductor.handlers import federation_conductor
from magnum import objects
from magnum.tests.unit.db import base as db_base
from magnum.tests.unit.db import utils


class TestHandler(db_base.DbTestCase):

    def setUp(self):
        super(TestHandler, self).setUp()
        self.handler = federation_conductor.Handler()
        federation_dict = utils.get_test_federation()
        self.federation = objects.Federation(self.context, **federation_dict)
        self.federation.create()

    def test_create_federation(self):
        self.assertRaises(NotImplementedError, self.handler.federation_create,
                          self.context, self.federation, create_timeout=15)

    def test_update_federation(self):
        self.assertRaises(NotImplementedError, self.handler.federation_update,
                          self.context, self.federation, rollback=False)

    def test_delete_federation(self):
        self.assertRaises(NotImplementedError, self.handler.federation_delete,
                          self.context, self.federation.uuid)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import subprocess

from tempest.lib import base

import magnum


COPY_LOG_HELPER = "magnum/tests/contrib/copy_instance_logs.sh"


class BaseMagnumTest(base.BaseTestCase):
    """Sets up configuration required for functional tests"""

    LOG = logging.getLogger(__name__)

    def __init__(self, *args, **kwargs):
        super(BaseMagnumTest, self).__init__(*args, **kwargs)

    @classmethod
    def copy_logs_handler(cls, get_nodes_fn, coe, keypair):
        """Copy logs closure.

        This method will retrieve all running nodes for a specified cluster
        and copy addresses from there locally.

        :param get_nodes_fn: function that takes no parameters and returns
            a list of node IPs which are in such form:
                [[master_nodes], [slave_nodes]].
        :param coe: the COE type of the nodes
        """
        def int_copy_logs():
            try:
                cls.LOG.info("Copying logs...")
                func_name = "test"
                msg = ("Failed to copy logs for cluster")
                nodes_addresses = get_nodes_fn()

                master_nodes = nodes_addresses[0]
                slave_nodes = nodes_addresses[1]

                base_path = os.path.split(os.path.dirname(
                    os.path.abspath(magnum.__file__)))[0]
                full_location = os.path.join(base_path, COPY_LOG_HELPER)

                def do_copy_logs(prefix, nodes_address):
                    if not nodes_address:
                        return

                    msg = "copy logs from : %s" % ','.join(nodes_address)
                    cls.LOG.info(msg)
                    log_name = prefix + "-" + func_name
                    for node_address in nodes_address:
                        try:
                            cls.LOG.debug("running %s", full_location)
                            cls.LOG.debug("keypair: %s", keypair)
                            subprocess.check_call([
                                full_location,
                                node_address,
                                coe,
                                log_name,
                                str(keypair)
                            ])
                        except Exception:
                            cls.LOG.error(msg)
                            msg = (
                                "failed to copy from %(node_address)s "
                                "to %(base_path)s%(log_name)s-"
                                "%(node_address)s" %
                                {'node_address': node_address,
                                 'base_path': "/opt/stack/logs/cluster-nodes/",
                                 'log_name': log_name})
                            cls.LOG.exception(msg)

                do_copy_logs('master', master_nodes)
                do_copy_logs('node', slave_nodes)
            except Exception:
                cls.LOG.exception(msg)

        return int_copy_logs



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from marathon import MarathonClient

from magnum.conductor.scale_manager import ScaleManager


class MesosScaleManager(ScaleManager):
    """When scaling a mesos cluster, MesosScaleManager will inspect the

    nodes and find out those with containers on them. Thus we can
    ask Heat to delete the nodes without containers. Note that this
    is a best effort basis -- Magnum doesn't have any synchronization
    with Marathon, so while Magnum is checking for the containers to
    choose nodes to remove, new containers can be deployed on the
    nodes to be removed.
    """

    def __init__(self, context, osclient, cluster):
        super(MesosScaleManager, self).__init__(context, osclient, cluster)

    def _get_hosts_with_container(self, context, cluster):
        marathon_client = MarathonClient(
            'http://' + cluster.api_address + ':8080')
        hosts = set()
        for task in marathon_client.list_tasks():
            hosts.add(task.host)

        return hosts



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tempest import clients
from tempest.common import credentials_factory as common_creds

from magnum.tests.functional.api.v1.clients import bay_client
from magnum.tests.functional.api.v1.clients import baymodel_client
from magnum.tests.functional.api.v1.clients import cert_client
from magnum.tests.functional.api.v1.clients import cluster_client
from magnum.tests.functional.api.v1.clients import cluster_template_client
from magnum.tests.functional.api.v1.clients import magnum_service_client
from magnum.tests.functional.common import client
from magnum.tests.functional.common import config


class Manager(clients.Manager):
    def __init__(self, credentials=None, request_type=None):
        if not credentials:
            credentials = common_creds.get_configured_credentials(
                'identity_admin')
        super(Manager, self).__init__(credentials)
        self.auth_provider.orig_base_url = self.auth_provider.base_url
        self.auth_provider.base_url = self.bypassed_base_url
        auth = self.auth_provider
        if request_type == 'baymodel':
            self.client = baymodel_client.BayModelClient(auth)
        elif request_type == 'bay':
            self.client = bay_client.BayClient(auth)
        elif request_type == 'cert':
            self.client = cert_client.CertClient(auth)
        elif request_type == 'cluster_template':
            self.client = cluster_template_client.ClusterTemplateClient(auth)
        elif request_type == 'cluster':
            self.client = cluster_client.ClusterClient(auth)
        elif request_type == 'service':
            self.client = magnum_service_client.MagnumServiceClient(auth)
        else:
            self.client = client.MagnumClient(auth)

    def bypassed_base_url(self, filters, auth_data=None):
        if (config.Config.magnum_url and
                filters['service'] == 'container-infra'):
            return config.Config.magnum_url
        return self.auth_provider.orig_base_url(filters, auth_data=auth_data)


class DefaultManager(Manager):
    def __init__(self, credentials, request_type=None):
        super(DefaultManager, self).__init__(credentials, request_type)


class AltManager(Manager):
    def __init__(self, credentials, request_type=None):
        super(AltManager, self).__init__(credentials, request_type)


class AdminManager(Manager):
    def __init__(self, credentials, request_type=None):
        super(AdminManager, self).__init__(credentials, request_type)



# Copyright 2016 Rackspace Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

import magnum.conf
from magnum.drivers.heat import k8s_fedora_template_def as kftd

CONF = magnum.conf.CONF


class AtomicK8sTemplateDefinition(kftd.K8sFedoraTemplateDefinition):
    """Kubernetes template for a Fedora Atomic VM."""

    @property
    def driver_module_path(self):
        return __name__[:__name__.rindex('.')]

    @property
    def template_path(self):
        return os.path.join(os.path.dirname(os.path.realpath(__file__)),
                            'templates/kubecluster.yaml')



# Copyright 2014
# The Cloudscaling Group, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from magnum.api import app as api_app
from magnum.api import config as api_config
from magnum.api import hooks
from magnum.tests import base


class TestAppConfig(base.BaseTestCase):

    def test_get_pecan_config(self):
        config = api_app.get_pecan_config()

        config_d = dict(config.app)

        self.assertEqual(api_config.app['modules'], config_d['modules'])
        self.assertEqual(api_config.app['root'], config_d['root'])
        self.assertIsInstance(config_d['hooks'][0], hooks.ContextHook)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import inspect

from magnum.common import exception
from magnum.i18n import _
from magnum.tests import base


class TestMagnumException(exception.MagnumException):
    message = _("templated %(name)s")


class TestException(base.BaseTestCase):

    def raise_(self, ex):
        raise ex

    def test_message_is_templated(self):
        ex = TestMagnumException(name="NAME")
        self.assertEqual("templated NAME", str(ex))

    def test_custom_message_is_templated(self):
        ex = TestMagnumException(_("custom templated %(name)s"), name="NAME")
        self.assertEqual("custom templated NAME", str(ex))

    def test_all_exceptions(self):
        for name, obj in inspect.getmembers(exception):
            if inspect.isclass(obj) and issubclass(obj, Exception):
                self.assertRaises(obj, self.raise_, obj())



# Copyright 2015 Huawei Technologies Co.,LTD.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import decorator

import pecan

from keystoneauth1 import exceptions as ka_exception

from magnum.api import utils as api_utils
from magnum.common import clients
from magnum.common import exception
import magnum.conf
from magnum.drivers.common import driver
from magnum.i18n import _
from magnum import objects

CONF = magnum.conf.CONF

cluster_update_allowed_properties = set(['node_count', 'health_status',
                                         'health_status_reason'])
federation_update_allowed_properties = set(['member_ids', 'properties'])


def ct_not_found_to_bad_request():
    @decorator.decorator
    def wrapper(func, *args, **kwargs):
        try:
            return func(*args, **kwargs)
        except exception.ClusterTemplateNotFound as e:
            # Change error code because 404 (NotFound) is inappropriate
            # response for a POST request to create a Cluster
            e.code = 400  # BadRequest
            raise

    return wrapper


def enforce_cluster_type_supported():
    @decorator.decorator
    def wrapper(func, *args, **kwargs):
        cluster = args[1]
        cluster_template = objects.ClusterTemplate.get(
            pecan.request.context, cluster.cluster_template_id)
        cluster_type = (cluster_template.server_type,
                        cluster_template.cluster_distro,
                        cluster_template.coe)
        driver.Driver.get_driver(*cluster_type)
        return func(*args, **kwargs)

    return wrapper


def enforce_driver_supported():
    @decorator.decorator
    def wrapper(func, *args, **kwargs):
        cluster_template = args[1]
        cluster_distro = cluster_template.cluster_distro
        if not cluster_distro:
            try:
                cli = clients.OpenStackClients(pecan.request.context)
                image_id = cluster_template.image_id
                image = api_utils.get_openstack_resource(cli.glance().images,
                                                         image_id,
                                                         'images')
                cluster_distro = image.get('os_distro')
            except Exception:
                pass
        cluster_type = (cluster_template.server_type,
                        cluster_distro,
                        cluster_template.coe)
        driver.Driver.get_driver(*cluster_type)
        return func(*args, **kwargs)

    return wrapper


def enforce_cluster_volume_storage_size():
    @decorator.decorator
    def wrapper(func, *args, **kwargs):
        cluster = args[1]
        cluster_template = objects.ClusterTemplate.get(
            pecan.request.context, cluster.cluster_template_id)
        _enforce_volume_storage_size(
            cluster_template.as_dict(), cluster.as_dict())
        return func(*args, **kwargs)

    return wrapper


def enforce_valid_project_id_on_create():
    @decorator.decorator
    def wrapper(func, *args, **kwargs):
        quota = args[1]
        _validate_project_id(quota.project_id)
        return func(*args, **kwargs)

    return wrapper


def _validate_project_id(project_id):
    try:
        context = pecan.request.context
        osc = clients.OpenStackClients(context)
        osc.keystone().domain_admin_client.projects.get(project_id)
    except ka_exception.http.NotFound:
        raise exception.ProjectNotFound(name='project_id',
                                        id=project_id)


def enforce_network_driver_types_create():
    @decorator.decorator
    def wrapper(func, *args, **kwargs):
        cluster_template = args[1]
        _enforce_network_driver_types(cluster_template)
        return func(*args, **kwargs)

    return wrapper


def enforce_network_driver_types_update():
    @decorator.decorator
    def wrapper(func, *args, **kwargs):
        cluster_template_ident = args[1]
        patch = args[2]
        cluster_template = api_utils.get_resource('ClusterTemplate',
                                                  cluster_template_ident)
        try:
            cluster_template_dict = api_utils.apply_jsonpatch(
                cluster_template.as_dict(), patch)
        except api_utils.JSONPATCH_EXCEPTIONS as e:
            raise exception.PatchError(patch=patch, reason=e)
        cluster_template = objects.ClusterTemplate(pecan.request.context,
                                                   **cluster_template_dict)
        _enforce_network_driver_types(cluster_template)
        return func(*args, **kwargs)

    return wrapper


def _enforce_network_driver_types(cluster_template):
    validator = Validator.get_coe_validator(cluster_template.coe)
    if not cluster_template.network_driver:
        cluster_template.network_driver = validator.default_network_driver
    validator.validate_network_driver(cluster_template.network_driver)


def enforce_server_type():
    @decorator.decorator
    def wrapper(func, *args, **kwargs):
        cluster_template = args[1]
        _enforce_server_type(cluster_template)
        return func(*args, **kwargs)

    return wrapper


def _enforce_server_type(cluster_template):
    validator = Validator.get_coe_validator(cluster_template.coe)
    validator.validate_server_type(cluster_template.server_type)


def enforce_volume_driver_types_create():
    @decorator.decorator
    def wrapper(func, *args, **kwargs):
        cluster_template = args[1]
        _enforce_volume_driver_types(cluster_template.as_dict())
        return func(*args, **kwargs)

    return wrapper


def enforce_volume_storage_size_create():
    @decorator.decorator
    def wrapper(func, *args, **kwargs):
        cluster_template = args[1]
        _enforce_volume_storage_size(cluster_template.as_dict(), {})
        return func(*args, **kwargs)

    return wrapper


def enforce_volume_driver_types_update():
    @decorator.decorator
    def wrapper(func, *args, **kwargs):
        cluster_template_ident = args[1]
        patch = args[2]
        cluster_template = api_utils.get_resource('ClusterTemplate',
                                                  cluster_template_ident)
        try:
            cluster_template_dict = api_utils.apply_jsonpatch(
                cluster_template.as_dict(), patch)
        except api_utils.JSONPATCH_EXCEPTIONS as e:
            raise exception.PatchError(patch=patch, reason=e)
        _enforce_volume_driver_types(cluster_template_dict)
        return func(*args, **kwargs)

    return wrapper


def _enforce_volume_driver_types(cluster_template):
    validator = Validator.get_coe_validator(cluster_template['coe'])
    if not cluster_template.get('volume_driver'):
        return
    validator.validate_volume_driver(cluster_template['volume_driver'])


def _enforce_volume_storage_size(cluster_template, cluster):
    volume_size = cluster.get('docker_volume_size') \
        or cluster_template.get('docker_volume_size')
    storage_driver = cluster_template.get('docker_storage_driver')

    if storage_driver == 'devicemapper':
        if not volume_size or volume_size < 3:
            raise exception.InvalidParameterValue(
                'docker volume size %s GB is not valid, '
                'expecting minimum value 3GB for %s storage '
                'driver.' % (volume_size, storage_driver))


def validate_cluster_properties(delta):

    update_disallowed_properties = delta - cluster_update_allowed_properties
    if update_disallowed_properties:
        err = (_("cannot change cluster property(ies) %s.") %
               ", ".join(update_disallowed_properties))
        raise exception.InvalidParameterValue(err=err)


def validate_federation_properties(delta):

    update_disallowed_properties = delta - federation_update_allowed_properties
    if update_disallowed_properties:
        err = (_("cannot change federation property(ies) %s.") %
               ", ".join(update_disallowed_properties))
        raise exception.InvalidParameterValue(err=err)


class Validator(object):

    @classmethod
    def get_coe_validator(cls, coe):
        if coe == 'kubernetes':
            return K8sValidator()
        elif coe == 'swarm' or coe == 'swarm-mode':
            return SwarmValidator()
        else:
            raise exception.InvalidParameterValue(
                _('Requested COE type %s is not supported.') % coe)

    @classmethod
    def validate_network_driver(cls, driver):
        cls._validate_network_driver_supported(driver)
        cls._validate_network_driver_allowed(driver)

    @classmethod
    def _validate_network_driver_supported(cls, driver):
        """Confirm that driver is supported by Magnum for this COE."""
        if driver not in cls.supported_network_drivers:
            raise exception.InvalidParameterValue(_(
                'Network driver type %(driver)s is not supported, '
                'expecting a %(supported_drivers)s network driver.') % {
                    'driver': driver,
                    'supported_drivers': '/'.join(
                        cls.supported_network_drivers + ['unspecified'])})

    @classmethod
    def _validate_network_driver_allowed(cls, driver):
        """Confirm that driver is allowed via configuration for this COE."""
        if ('all' not in cls.allowed_network_drivers and
           driver not in cls.allowed_network_drivers):
            raise exception.InvalidParameterValue(_(
                'Network driver type %(driver)s is not allowed, '
                'expecting a %(allowed_drivers)s network driver. ') % {
                    'driver': driver,
                    'allowed_drivers': '/'.join(
                        cls.allowed_network_drivers + ['unspecified'])})

    @classmethod
    def validate_volume_driver(cls, driver):
        cls._validate_volume_driver_supported(driver)

    @classmethod
    def _validate_volume_driver_supported(cls, driver):
        """Confirm that volume driver is supported by Magnum for this COE."""
        if driver not in cls.supported_volume_driver:
            raise exception.InvalidParameterValue(_(
                'Volume driver type %(driver)s is not supported, '
                'expecting a %(supported_volume_driver)s volume driver.') % {
                    'driver': driver,
                    'supported_volume_driver': '/'.join(
                        cls.supported_volume_driver + ['unspecified'])})

    @classmethod
    def validate_server_type(cls, server_type):
        cls._validate_server_type(server_type)

    @classmethod
    def _validate_server_type(cls, server_type):
        """Confirm that server type is supported by Magnum for this COE."""
        if server_type not in cls.supported_server_types:
            raise exception.InvalidParameterValue(_(
                'Server type %(server_type)s is not supported, '
                'expecting a %(supported_server_types)s server type.') % {
                    'server_type': server_type,
                    'supported_server_types': '/'.join(
                        cls.supported_server_types + ['unspecified'])})


class K8sValidator(Validator):

    supported_network_drivers = ['flannel', 'calico']
    supported_server_types = ['vm', 'bm']
    allowed_network_drivers = (
        CONF.cluster_template.kubernetes_allowed_network_drivers)
    default_network_driver = (
        CONF.cluster_template.kubernetes_default_network_driver)

    supported_volume_driver = ['cinder']


class SwarmValidator(Validator):

    supported_network_drivers = ['docker', 'flannel']
    supported_server_types = ['vm', 'bm']
    allowed_network_drivers = (CONF.cluster_template.
                               swarm_allowed_network_drivers)
    default_network_driver = (CONF.cluster_template.
                              swarm_default_network_driver)

    supported_volume_driver = ['rexray']


class MesosValidator(Validator):

    supported_network_drivers = ['docker']
    supported_server_types = ['vm', 'bm']
    allowed_network_drivers = (CONF.cluster_template.
                               mesos_allowed_network_drivers)
    default_network_driver = (CONF.cluster_template.
                              mesos_default_network_driver)

    supported_volume_driver = ['rexray']



# Copyright 2015 Huawei Technologies Co.,LTD.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests

from magnum.conductor.handlers.common.cert_manager import create_client_files


class KubernetesAPI:
    """
    Simple Kubernetes API client using requests.

    This API wrapper allows for a set of very simple operations to be
    performed on a Kubernetes cluster using the `requests` library. The
    reason behind it is that the native `kubernetes` library does not
    seem to be quite thread-safe at the moment.

    Also, our interactions with the Kubernetes API are happening inside
    Greenthreads so we don't need to use connection pooling on top of it,
    in addition to pools not being something that you can disable with
    the native Kubernetes API.
    """

    def __init__(self, context, cluster):
        self.context = context
        self.cluster = cluster

        # Load certificates for cluster
        (self.ca_file, self.key_file, self.cert_file) = create_client_files(
            self.cluster, self.context
        )

    def _request(self, method, url, json=True):
        response = requests.request(
            method,
            url,
            verify=self.ca_file.name,
            cert=(self.cert_file.name, self.key_file.name)
        )
        response.raise_for_status()
        if json:
            return response.json()
        else:
            return response.text

    def get_healthz(self):
        """
        Get the health of the cluster from API
        """
        return self._request(
            'GET',
            f"{self.cluster.api_address}/healthz",
            json=False
        )

    def list_node(self):
        """
        List all nodes in the cluster.

        :return: List of nodes.
        """
        return self._request(
            'GET',
            f"{self.cluster.api_address}/api/v1/nodes"
        )

    def list_namespaced_pod(self, namespace):
        """
        List all pods in the given namespace.

        :param namespace: Namespace to list pods from.
        :return: List of pods.
        """
        return self._request(
            'GET',
            f"{self.cluster.api_address}/api/v1/namespaces/{namespace}/pods"
        )

    def __del__(self):
        """
        Close all of the file descriptions for the certificates, since they
        are left open by `create_client_files`.

        TODO(mnaser): Use a context manager and avoid having these here.
        """
        if hasattr(self, 'ca_file'):
            self.ca_file.close()
        if hasattr(self, 'cert_file'):
            self.cert_file.close()
        if hasattr(self, 'key_file'):
            self.key_file.close()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields

from magnum.db import api as dbapi
from magnum.objects import base


@base.MagnumObjectRegistry.register
class Quota(base.MagnumPersistentObject, base.MagnumObject,
            base.MagnumObjectDictCompat):
    # Version 1.0: Initial version
    VERSION = '1.0'

    dbapi = dbapi.get_instance()

    fields = {
        'id': fields.IntegerField(),
        'project_id': fields.StringField(nullable=False),
        'resource': fields.StringField(nullable=False),
        'hard_limit': fields.IntegerField(nullable=False),
    }

    @base.remotable_classmethod
    def get_quota_by_project_id_resource(cls, context, project_id, resource):
        """Find a quota based on its integer id and return a Quota object.

        :param project_id: the id of a project.
        :param resource: resource name.
        :param context: Security context
        :returns: a :class:`Quota` object.
        """
        db_quota = cls.dbapi.get_quota_by_project_id_resource(project_id,
                                                              resource)
        quota = Quota._from_db_object(cls(context), db_quota)
        return quota

    @staticmethod
    def _from_db_object(quota, db_quota):
        """Converts a database entity to a formal object."""
        for field in quota.fields:
            setattr(quota, field, db_quota[field])

        quota.obj_reset_changes()
        return quota

    @staticmethod
    def _from_db_object_list(db_objects, cls, context):
        """Converts a list of database entities to a list of formal objects."""
        return [Quota._from_db_object(cls(context), obj)
                for obj in db_objects]

    @base.remotable_classmethod
    def get_by_id(cls, context, quota_id):
        """Find a quota based on its integer id and return a Quota object.

        :param quota_id: the id of a quota.
        :param context: Security context
        :returns: a :class:`Quota` object.
        """
        db_quota = cls.dbapi.get_quota_by_id(context, quota_id)
        quota = Quota._from_db_object(cls(context), db_quota)
        return quota

    @base.remotable_classmethod
    def list(cls, context, limit=None, marker=None,
             sort_key=None, sort_dir=None, filters=None):
        """Return a list of Quota objects.

        :param context: Security context.
        :param limit: maximum number of resources to return in a single result.
        :param marker: pagination marker for large data sets.
        :param sort_key: column to sort results by.
        :param sort_dir: direction to sort. "asc" or "desc".
        :param filters: filter dict, can includes 'project_id',
                        'resource'.
        :returns: a list of :class:`Quota` object.

        """
        db_quotas = cls.dbapi.get_quota_list(context,
                                             limit=limit,
                                             marker=marker,
                                             sort_key=sort_key,
                                             sort_dir=sort_dir,
                                             filters=filters)
        return Quota._from_db_object_list(db_quotas, cls, context)

    @base.remotable_classmethod
    def quota_get_all_by_project_id(cls, context, project_id):
        """Find a quota based on project id.

        :param project_id: the project id.
        :param context: Security context
        :returns: a :class:`Quota` object.
        """
        quotas = cls.dbapi.get_quota_by_project_id(context, project_id)
        return Quota._from_db_object_list(quotas, cls, context)

    @base.remotable
    def create(self, context=None):
        """Save a quota based on project id.

        :param context: security context.
        :returns: a :class:`Quota` object.
        """
        values = self.obj_get_changes()
        db_quota = self.dbapi.create_quota(values)
        self._from_db_object(self, db_quota)

    @base.remotable
    def delete(self, context=None):
        """Delete the quota from the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: Quota(context)
        """
        self.dbapi.delete_quota(self.project_id, self.resource)
        self.obj_reset_changes()

    @base.remotable_classmethod
    def update_quota(cls, context, project_id, quota):
        """Save a quota based on project id.

        :param quota: quota.
        :returns: a :class:`Quota` object.
        """
        db_quota = cls.dbapi.update_quota(project_id, quota)
        return Quota._from_db_object(cls(context), db_quota)



# Copyright 2015 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from heatclient import exc
from oslo_log import log as logging
from pycadf import cadftaxonomy as taxonomy

from magnum.common import exception
from magnum.common import profiler
from magnum.conductor.handlers.common import cert_manager
from magnum.conductor import utils as conductor_utils
from magnum.drivers.common import driver
from magnum.i18n import _
from magnum import objects
from magnum.objects import fields

import six
LOG = logging.getLogger(__name__)


@profiler.trace_cls("rpc")
class Handler(object):
    """Magnum CA RPC handler.

    These are the backend operations. They are executed by the backend service.
    API calls via AMQP (within the ReST API) trigger the handlers to be called.

    """

    def __init__(self):
        super(Handler, self).__init__()

    def sign_certificate(self, context, cluster, certificate):
        LOG.debug("Creating self signed x509 certificate")
        try:
            ca_cert_type = certificate.ca_cert_type
        except Exception as e:
            LOG.debug("There is no CA cert type specified for the CSR")
            ca_cert_type = "kubernetes"

        signed_cert = cert_manager.sign_node_certificate(cluster,
                                                         certificate.csr,
                                                         ca_cert_type,
                                                         context=context)
        if six.PY3 and isinstance(signed_cert, six.binary_type):
            certificate.pem = signed_cert.decode()
        else:
            certificate.pem = signed_cert
        return certificate

    def get_ca_certificate(self, context, cluster, ca_cert_type=None):
        ca_cert = cert_manager.get_cluster_ca_certificate(
            cluster, context=context, ca_cert_type=ca_cert_type)
        certificate = objects.Certificate.from_object_cluster(cluster)
        if six.PY3 and isinstance(ca_cert.get_certificate(), six.binary_type):
            certificate.pem = ca_cert.get_certificate().decode()
        else:
            certificate.pem = ca_cert.get_certificate()
        return certificate

    def rotate_ca_certificate(self, context, cluster):
        LOG.info('start rotate_ca_certificate for cluster: %s', cluster.uuid)

        allow_update_status = (
            fields.ClusterStatus.CREATE_COMPLETE,
            fields.ClusterStatus.UPDATE_COMPLETE,
            fields.ClusterStatus.RESUME_COMPLETE,
            fields.ClusterStatus.RESTORE_COMPLETE,
            fields.ClusterStatus.ROLLBACK_COMPLETE,
            fields.ClusterStatus.SNAPSHOT_COMPLETE,
            fields.ClusterStatus.CHECK_COMPLETE,
            fields.ClusterStatus.ADOPT_COMPLETE
        )
        if cluster.status not in allow_update_status:
            conductor_utils.notify_about_cluster_operation(
                context, taxonomy.ACTION_UPDATE, taxonomy.OUTCOME_FAILURE,
                cluster)
            operation = _('Updating a cluster when status is '
                          '"%s"') % cluster.status
            raise exception.NotSupported(operation=operation)

        try:
            # re-generate the ca certs
            cert_manager.generate_certificates_to_cluster(cluster,
                                                          context=context)
            cluster_driver = driver.Driver.get_driver_for_cluster(context,
                                                                  cluster)
            cluster_driver.rotate_ca_certificate(context, cluster)
            cluster.status = fields.ClusterStatus.UPDATE_IN_PROGRESS
            cluster.status_reason = None
        except Exception as e:
            cluster.status = fields.ClusterStatus.UPDATE_FAILED
            cluster.status_reason = six.text_type(e)
            cluster.save()
            conductor_utils.notify_about_cluster_operation(
                context, taxonomy.ACTION_UPDATE, taxonomy.OUTCOME_FAILURE,
                cluster)
            if isinstance(e, exc.HTTPBadRequest):
                e = exception.InvalidParameterValue(message=six.text_type(e))
                raise e
            raise

        cluster.save()
        return cluster



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_policy import policy

from magnum.common.policies import base

CERTIFICATE = 'certificate:%s'
RULE_ADMIN_OR_USER_OR_CLUSTER_USER = base.RULE_ADMIN_OR_USER + " or " + \
    base.RULE_CLUSTER_USER

rules = [
    policy.DocumentedRuleDefault(
        name=CERTIFICATE % 'create',
        check_str=RULE_ADMIN_OR_USER_OR_CLUSTER_USER,
        description='Sign a new certificate by the CA.',
        operations=[
            {
                'path': '/v1/certificates',
                'method': 'POST'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=CERTIFICATE % 'get',
        check_str=RULE_ADMIN_OR_USER_OR_CLUSTER_USER,
        description='Retrieve CA information about the given bay/cluster.',
        operations=[
            {
                'path': '/v1/certificates/{bay_uuid/cluster_uuid}',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=CERTIFICATE % 'rotate_ca',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description='Rotate the CA certificate on the given bay/cluster.',
        operations=[
            {
                'path': '/v1/certificates/{bay_uuid/cluster_uuid}',
                'method': 'PATCH'
            }
        ]
    )
]


def list_rules():
    return rules
